/***************************************************************************************************
*             C# sample for the usage of Episodic TopoART (class Fast_Episodic_TopoART)            *
****************************************************************************************************
*                             Created by Marko Tscherepanow, 27 May 2016                           *
***************************************************************************************************/

// Compile and run from the console: dotnet run --project Episodic_TopoART_sample2.csproj

using System;
using System.IO;
using System.Reflection;
using SixLabors.ImageSharp;
using SixLabors.ImageSharp.PixelFormats;
using System.Runtime.InteropServices;
using LibTopoART;

namespace LibTopoART_samples
{
	/// <summary>
	/// Episodic clustering sample using real-world video data. [C#]
	/// <para>
	/// Like in Section 4.2 of "Marko Tscherepanow, Sina Kühnel, and Sören Riechers (2012). Episodic
	/// Clustering of Data Streams Using a Topology-Learning Neural Network. In Proceedings of the European 
	/// Conference on Artificial Intelligence (ECAI), Workshop on Active and Incremental Learning (AIL), 
	/// pp. 24-29. Montpellier, France.", an Episodic TopoART network is trained with real-world video data. 
	/// Each image has a size of 64x36 pixels. As each pixel comprises 3 color channels (RGB), the input 
	/// length equals 6912. After finishing training, recall is performed for a single input stimulus.
	/// </para>
	/// <para>The recall results can be visualised using the script <c>ShowEpisodicTopoARTRecallResults</c> 
	/// provided for R and MATLAB in the subfolder <c>visualisation</c>.
	/// </para>
	/// </summary>
	class Episodic_TopoART_sample2
	{
		static void Main() {
			
			// Dataset (containing training and test images)
			const string DatasetPath	=	"../../../../../data/AIL12-like_video_dataset/";
			// Destination directory for trained networks
			const string NetworkPath	=	"../../../../../results/networks/";
			const long SampleNumber		=	59523;

			// Recall stimulus from the test set (only odd images allowed)
			const string RecallStimulus	=	"image_54971.jpg";
			// Destination directory for recall results
			const string RecallPath		=	"../../../../../results/recall/AIL12-like_video_dataset_recall_results/";
			// Recall control parameters
			const long MaxInterEpisodeRecallSteps			=	15;
			const long MaxIntraEpisodeRecallSteps			=	50;
			const decimal MinInterEpisodeRecallActivation	=	0.5m;

			// Set to true to omit training and use a saved network file
			bool UseSavedNetwork = false;

			byte[] InputArray;
			int i, j;
			int width, height;
			TimeSpan time;
			Fast_Episodic_TopoART feta = null;
			string next_image;

			// Set working directory to assembly directory
			Directory.SetCurrentDirectory(Path.GetDirectoryName(new Uri(Assembly.GetEntryAssembly().Location).LocalPath));

			// Load dataset and train a new network
			if(!UseSavedNetwork) {

				// Start time measuring
				DateTime training_start = DateTime.Now;

				// Get image size and first image
				i = 0;
				LoadImage(DatasetPath + "train/image_" + i.ToString("D5") + ".jpg", out width, out height, out InputArray);
				feta			= 	new Fast_Episodic_TopoART(InputArray.Length, 2, 0.7m, 400);
				feta.Phi		=	5;
				feta.Tau		=	200;
				feta.Beta_sbm	=	0.25m;

				do {
					if(i != 0) {
						LoadImage(DatasetPath + "train/image_" + i.ToString("D5") + ".jpg", out width, out height, out InputArray);
						if((i % 1000) == 0)
							Console.Write(".");
					}
					feta.Learn(InputArray);
					i += 2;
				} while(i < SampleNumber);

				Console.WriteLine(" ");

				// Stop time measuring
				DateTime training_end = DateTime.Now;

				// Output the required time
				time = training_end - training_start;
				Console.WriteLine("Time for training: " + time);

				Console.WriteLine("Save network");

				// Save network in human readable form
				// feta.SaveText(NetworkPath + "Fast_Episodic_TopoART_AIL12-like_video_dataset.txt");

				// Save network in binary form
				feta.Save(NetworkPath + "Fast_Episodic_TopoART_AIL12-like_video_dataset.feta");
			}
			// Load an existing network
			else {
				feta = new Fast_Episodic_TopoART(NetworkPath + "Fast_Episodic_TopoART_AIL12-like_video_dataset.eta");
			}

			// Start time measuring
			DateTime recall_start = DateTime.Now;

			// Perform recall
			if(feta != null) {
				// Copy stimulus
				File.Copy(DatasetPath + "test/" + RecallStimulus, RecallPath + "stimulus.jpg", true);

				// Load recall stimulus
				LoadImage(DatasetPath + "test/" + RecallStimulus, out width, out height, out InputArray);

				byte[] RecallResult;
				decimal activation;
				feta.BeginRecall(InputArray);
				feta.InterEpisodeRecallStep(out RecallResult, out activation);

				// Inter-episode recall loop
				for(i = 1; (RecallResult != null) && (activation > MinInterEpisodeRecallActivation) && (i <= MaxInterEpisodeRecallSteps); ++i) {
					SaveImage(RecallPath + "inter_image_" + i.ToString("D3") + ".jpg", width, height, RecallResult);

					// Intra-episode recall loop
					for(j = 1; (RecallResult != null) && (j <= MaxIntraEpisodeRecallSteps); ++j) {
						feta.IntraEpisodeRecallStep(out RecallResult);
						if(RecallResult == null) 
							break;
						SaveImage(RecallPath + "intra_image_" + i.ToString("D3") + "_" + j.ToString("D3") + ".jpg", width, height, RecallResult);
					}

					// Signify stopped intra-episode recall
					next_image = RecallPath + "intra_image_" + i.ToString("D3") + "_" + j.ToString("D3") + ".jpg";
					if(File.Exists(next_image))
						File.Delete(next_image);

					feta.InterEpisodeRecallStep(out RecallResult, out activation);
				}

				// Signify stopped inter-episode recall
				next_image = RecallPath + "inter_image_" + i.ToString("D3") + ".jpg";
				if(File.Exists(next_image))
					File.Delete(next_image);

				feta.EndRecall();
			}

			// Stop time measuring
			DateTime recall_end = DateTime.Now;

			// Output the required time
			time = recall_end - recall_start;
			Console.WriteLine("Time for recall: " + time);
		}

		private static void LoadImage(string path, out int width, out int height, out byte[] ImageArray) {
			using var image = Image.Load<Rgb24>(path);
			width = image.Width;
			height = image.Height;
			Span<Rgb24> data = new Span<Rgb24>(new Rgb24[image.Width * image.Height]);
			image.CopyPixelDataTo(data);
			ImageArray = MemoryMarshal.AsBytes(data).ToArray();
		}

		private static void SaveImage(string path, int width, int height, byte[] ImageArray) {
			using var image = new Image<Rgb24>(width, height);
			for(long i = 0, j = 0; j < ImageArray.LongLength; ++i, j += 3)
				image[(int)(i % width), (int)(i / width)] = new Rgb24(ImageArray[j], ImageArray[j + 1], ImageArray[j + 2]);
			image.SaveAsJpeg(path);
		}
	}
}