/***************************************************************************************************
*           C# sample for the usage of TopoART-R (classes TopoART_R and Fast_TopoART_R)            *
*                                  (approximation of a sine function)                              *
****************************************************************************************************
*                           Created by Marko Tscherepanow, 23 November 2016                        *
***************************************************************************************************/

// Compile and run from the console: dotnet run --project TopoART-R_sample2.csproj

using System;
using System.IO;
using System.Reflection;
using LibTopoART;

namespace LibTopoART_samples
{
	/// <summary>
	/// Regression sample using TopoART-R. (advanced version) [C#]
	/// <para>
	/// This sample trains a TopoART-R network with 100 points sampled from a sine function.
	/// Then, sine values are predicted for 25 random values.
	/// </para>
	/// <para>The predicted results can be visualised using the script <c>ShowTopoARTRResults</c> provided 
	/// for R and MATLAB in the subfolder <c>visualisation</c>.
	/// </para>
	/// </summary>
	class TopoART_R_sample2
	{
		enum Selector {
			USE_TopoART_R,
			USE_Fast_TopoART_R,
		};

		static void Main() {
			long i;

			// Destination directory for the regression results
			const string ResultPath = "../../../../../results/regression/";

			// Input and output dimensions
			const long i_len = 2;	// number of independent variables (input dimensions)
			const long d_len = 1;	// number of dependent variables (output dimensions)

			// Choose TopoART-R implementation (see enum Selector above)
			Selector sel = Selector.USE_Fast_TopoART_R;

			// Training data
			const long training_sample_num	= 100; 
			const long test_sample_num		= 25; 
			decimal[][] i_values = new decimal[training_sample_num][];
			decimal[][] d_values = new decimal[training_sample_num][];
					
			Random rnd = new Random();

			// Set working directory to assembly directory
			Directory.SetCurrentDirectory(Path.GetDirectoryName(new Uri(Assembly.GetEntryAssembly().Location).LocalPath));

			// Create training samples (sine function)
			for(i = 0; i < training_sample_num; ++i) {

				// Set the independent variables
				i_values[i]		=	new decimal[i_len];
				i_values[i][0]	=	(decimal)i * (decimal)Math.PI * 2.0m / (decimal)(training_sample_num - 1);

				// Simulate a second independent variable
				i_values[i][1]	=	(decimal)rnd.NextDouble();									

				// Set the dependent variable
				d_values[i]		=	new decimal[d_len];
				d_values[i][0]	=	((decimal)Math.Sin((double)i_values[i][0]) + 1.0m) / 2.0m; 

				// Normalisation of the first independent variable
				i_values[i][0]	/=	((decimal)Math.PI * 2.0m);									
			}

			// Create TopoART-R network using rho_a = 0.99 and default values for the remaining parameters
			ITopoART_R tar;
			if(sel == Selector.USE_TopoART_R)
				tar = new TopoART_R(i_len, d_len, 2, 0.99m);
			else
				tar = new Fast_TopoART_R(i_len, d_len, 2, 0.99m);

			// Create file for the results and write the training data
			// The second independent variable is not saved in oder to be compatible with the results of TopoART-R_sample1.
			using(TextWriter writer = new StreamWriter(File.Open(ResultPath + "TopoART-R_sine_regression_results.csv", 
				FileMode.Create)))
			{
				writer.WriteLine("   {0,6:D},   {1,6:D}", training_sample_num, test_sample_num);

				// Training with separated arrays for the independent variables and the dependent variables
				Console.WriteLine("Train TopoART-R network with {0} samples", training_sample_num);

				// Train the network until it stabilises and output the number of required iterations
				long iteration_num = 0;
				do {
					tar.ResetAdaptationState();

					for(i = 0; i < training_sample_num; ++i) {
						tar.Learn(i_values[i], d_values[i]);

						// Write denormalised training data (only in the first iteration)
						if(iteration_num == 0)
							writer.WriteLine("   {0},   {1}", 
								(i_values[i][0] * (decimal)Math.PI * 2.0m).ToString(" 0.000;-0.000", System.Globalization.CultureInfo.InvariantCulture), 
								((d_values[i][0] - 0.5m) * 2.0m).ToString(" 0.000;-0.000", System.Globalization.CultureInfo.InvariantCulture));
					}

					++iteration_num;
				} while((tar.GetAdaptationState() & AdaptationState.ANY_PERMANENT_ADAPTATION_MASK) != AdaptationState.NO_ADAPTATION);
				Console.WriteLine("Stopped training after {0} iterations", iteration_num);

				// Prediction with a mask vector (only one independent variable is present) 
				decimal[] random_i_vec = new decimal[i_len];
				TopoART_R_prediction<decimal> prediction;
				bool[] m_i_vec 	= new bool[i_len];
				m_i_vec[0] 		= false;				// first independent variable is presented
				m_i_vec[1] 		= true;					// second independent variable is missing
				Console.WriteLine("Predict sine function for 25 random values");
				for(i = 0; i < test_sample_num; ++i) {
					random_i_vec[0] = (decimal)rnd.NextDouble();
					prediction		= tar.Predict(random_i_vec, m_i_vec);
					writer.WriteLine("   {0},   {1}",
						(random_i_vec[0] * (decimal)Math.PI * 2.0m).ToString(" 0.000;-0.000", System.Globalization.CultureInfo.InvariantCulture), 
						((prediction.d_vec_prediction[0] - 0.5m) * 2.0m).ToString(" 0.000;-0.000", System.Globalization.CultureInfo.InvariantCulture)); 
				}
			}
		}
	}
}